// Computation of the first part of Lemma 2.2.

// The abstract setting is the following: we will have 'callers' which will call
// 'quadratures', which will call 'integrands'. Different callers decide for example
// when and how to split the integration domains in a different way, the quadratures
// implement different integration methods and the integrands are the different
// functions under the integral sign.

// We validate \pa_t \pa_x z^{1}(0,0) for two values of A(\eps), \eps = [0,1e-6],
// in which the sign change establishes the existence of a function A(\eps) for 
// which \pa_t \pa_x z^{1}(0,0) = 0.

// The previous quantity consists of three 1D integrals: each of them we split it
// into singular and non-singular regions.

// The singular regions are calculated expanding both the numerator
// and the denominator in Taylor series to avoid fractions of the type 0/0.
// The non-singular ones are calculated using a Gauss-Legendre quadrature of
// order 2. All integrations are done adaptively, bisecting the region of
// integration whenever some tolerances are not met, up to a maximum number of
// bisections. The bisection is done by the arithmetic mean of the endpoints.

// The program is split in 3(x2) more files:

// curves.h/cpp contains information about the family of curves and its
// implementation with the different data structures (interval,itaylor).
// It also contains the methods and the definition of the class ParameterSet,
// which will enclose all the information we need to perform the calculations.

// integrands.h/cpp contains the different integrands of the functions to be computed.

// methods.h/cpp contains both the caller and the quadrature methods.

// All calculations are done using rigorous arithmetics using the C-XSC library.

// Usage: ./Lemma_221

// Output (to the screen):

// 1.08050000000  1.08050000000 -0.00000000000  0.00000100000 [ 0.00001996941, 0.00026671564]
// 1.08055000000  1.08055000000 -0.00000000000  0.00000100000 [-0.00027111048,-0.00002436699]


#include "interval.hpp"
#include "itaylor.hpp"
#include "dim2taylor.hpp"
#include "curves.h"
#include "integrands.h"
#include "methods.h"

#include <iostream>
#include <string>
#include <queue>
#include <fstream>
#include <cassert>
#include <vector>

using namespace cxsc;
using namespace std;
using namespace taylor;



///////////////////////////////////////////////////////////
// BEGIN DEFINITION OF THE INTEGRALS TO BE COMPUTED
///////////////////////////////////////////////////////////

// Computes \pa_t \pa_x z^{1}(0,0) for two values of A(\eps), \eps = [0,1e-6].
// We are only interested in giving a sign out of each of the two values.
// With this function we show the existence of a function A(\eps) for which
// \pa_t \pa_x z^{1}(0,0) = 0.

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_221(const ParameterSet& p, int& flag){

  interval sing(0.0);
  interval res(0.0);
  
  // rec_level controls how deep we go into the division of the integrating region
  // and its depth is limited by constants specified in 'methods.h'
  int rec_level = 0; 


  // We start computing the bounded part
  rec_level++;
  res+=rec_integrate_1D(p,p.Left,p.Right,GaussLegendre2_1D,integrand_A1_fast,integrand_A1_slow,flag,rec_level);
  
  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }
  
  rec_level++;
  res+=rec_integrate_1D(p,p.Left,p.Right,GaussLegendre2_1D,integrand_A2_fast,integrand_A2_slow,flag,rec_level);
  
  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }
  
  rec_level++;
  res+=rec_integrate_1D(p,p.Left,p.Right,GaussLegendre2_1D,integrand_A3_fast,integrand_A3_slow,flag,rec_level);
  
  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }

  // We now compute the singular part
  rec_level++;
  sing += rec_integrate_singularity_1D(
                        p,(interval)0.0,p.Left,
                        integrate_singularity_1D,
                        integrand_A1_singularity_num_slow,integrand_A1_singularity_den_slow,
                        integrand_A1_fast,integrand_A1_slow,
                        flag,rec_level,2,2);

  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }
  
  rec_level++;
  sing += rec_integrate_singularity_1D(
                        p,(interval)0.0,p.Left,
                        integrate_singularity_1D,
                        integrand_A2_singularity_num_slow,integrand_A2_singularity_den_slow,
                        integrand_A2_fast,integrand_A2_slow,
                        flag,rec_level,2,2);

  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }
  
  rec_level++;
  sing += rec_integrate_singularity_1D(
                        p,(interval)0.0,p.Left,
                        integrate_singularity_1D,
                        integrand_A3_singularity_num_slow,integrand_A3_singularity_den_slow,
                        integrand_A3_fast,integrand_A3_slow,
                        flag,rec_level,4,4);

  assert(rec_level == 0);
  // Stop if division by zero
  if (flag & 1){
    return interval(0.0,0.0); 
  }
  
  return (res+sing);
}


int main(int argc, char* argv[]){
  cout << SetPrecision(14,11);
  if (argc != 1) {
    cout << "Usage: " << endl;
    cout << argv[0] << endl;
    cout << "(and some patience...)" << endl;
    cout << "Results will be output to screen as soon as they are computed." << endl;
    exit(0);
  }
  
  ParameterSet PS_ini;
  // Neighbourhood of the singularity set to 2^(-9)
  "[0.001953125,0.001953125]" >> PS_ini.Left;
  // Boundary for the bounded part of the second interval
  PS_ini.Right = Pi_interval;
  PS_ini.abs_tol = PS_ini.rel_tol = 1e-6;
  "[1.08050,1.08050]" >> PS_ini.A;
  "[0.000000,0.000001]" >> PS_ini.epsilon;
  
  int flag = 0;

  interval result_cell = compute_Lemma_221(PS_ini, flag);
  
  if ((flag&1)){ // critical
    cerr << "Critical case: division by zero encountered" << endl;
    return 0;
  }
  
  cerr << Inf(PS_ini.A) << " " << Sup(PS_ini.A) << " " 
       << Inf(PS_ini.epsilon) << " " << Sup(PS_ini.epsilon)
       << " " << result_cell << endl;

  "[1.08055,1.08055]" >> PS_ini.A;
  
  result_cell = compute_Lemma_221(PS_ini, flag);
  
  if ((flag&1)){ // critical
    cerr << "Critical case: division by zero encountered" << endl;
    return 0;
  }
  
  cerr << Inf(PS_ini.A) << " " << Sup(PS_ini.A) << " " 
       << Inf(PS_ini.epsilon) << " " << Sup(PS_ini.epsilon)
       << " " << result_cell << endl;


  return 0;
}

