#include "integrands.h"


// Integrand of A_{1}: itaylor data type
itaylor integrand_A1_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order+2,x);
  itaylor Z1 = z1(x_it,p);
  itaylor D2Z1 = diff(Z1, 2);
  Z1 = trunc(Z1,order);
  x_it = trunc(x_it,order);
  itaylor Z2 = z2(x_it,p);
  // Now, Z1 and D2Z1 are 'order'-order itaylor objects
  itaylor den = cosh(Z2)-cos(Z1);
  if (0 <= get_j_coef(den,0)){
    // Division by zero
    flag |= 1;
    return itaylor(0.0);
  }
  return  (interval)2.0*D2Z1*sin(Z1)/den;
}


// Integrand of A_{1}: interval data type
interval integrand_A1_fast(const interval& x, const ParameterSet& p, int& flag){
  itaylor x_it = itaylor(2,x);
  itaylor ITZ1 = z1(x_it,p);
  interval Z1 = get_j_derive(ITZ1,0);
  interval D2Z1 = get_j_derive(ITZ1,2);
  interval Z2 = z2(x,p);
  // Now, Z1 and D2Z1 are intervals
  interval den = cosh(Z2)-cos(Z1);
  if (0.0 <= den){
    // Division by zero
    flag |= 1;
    return interval(0.0);
  }
  return (interval)2.0*D2Z1*sin(Z1)/den;
}

// Integrand of A_{2}: itaylor data type
itaylor integrand_A2_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order+1,x);
  itaylor Z1 = z1(x_it,p);
  itaylor DZ1 = diff(Z1, 1);
  Z1 = trunc(Z1,order);
  x_it = trunc(x_it,order);
  
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  
  itaylor Z2 = z2(x_it,p);
  itaylor den = cosh(Z2)-cos(Z1);
  if (0 <= get_j_coef(den,0)){
    // Division by zero
    flag |= 1;
    return itaylor(0.0);
  }
  return  (interval)2.0*sqr(DZ10-DZ1)*cos(Z1)/den;
}


// Integrand of A_{2}: interval data type
interval integrand_A2_fast(const interval& x, const ParameterSet& p, int& flag){
  itaylor x_it = itaylor(1,x);
  itaylor ITZ1 = z1(x_it,p);
  interval Z1 = get_j_derive(ITZ1,0);
  interval DZ1 = get_j_derive(ITZ1,1);
  
  // Now, Z1 and DZ1 are intervals
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  
  interval Z2 = z2(x,p);
  interval den = cosh(Z2)-cos(Z1);
  if (0.0 <= den){
    // Division by zero
    flag |= 1;
    return interval(0.0);
  }
  return  (interval)2.0*sqr(DZ10-DZ1)*cos(Z1)/den;
}

// Integrand of A_{3}: itaylor data type
itaylor integrand_A3_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order+1,x);
  itaylor Z1 = z1(x_it,p);
  itaylor Z2 = z2(x_it,p);
  itaylor DZ1 = diff(Z1, 1);
  itaylor DZ2 = diff(Z2, 1);
  Z1 = trunc(Z1,order);
  Z2 = trunc(Z2,order);
  x_it = trunc(x_it,order);
  
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  itaylor ITZ20 = z2(x_it0,p);
  interval DZ20 = get_j_derive(ITZ20,1);
  
  itaylor den = cosh(Z2)-cos(Z1);
  if (0 <= get_j_coef(den,0)){
    // Division by zero
    flag |= 1;
    return itaylor(0.0);
  }
  return  (interval)(-2.0)*(DZ10-DZ1)*sin(Z1)*(sinh(Z2)*(DZ20-DZ2) + sin(Z1)*(DZ10-DZ1))/sqr(den);
}


// Integrand of A_{3}: interval data type
interval integrand_A3_fast(const interval& x, const ParameterSet& p, int& flag){
  itaylor x_it = itaylor(1,x);
  itaylor ITZ1 = z1(x_it,p);
  itaylor ITZ2 = z2(x_it,p);
  interval Z1 = get_j_coef(ITZ1,0);
  interval DZ1 = get_j_coef(ITZ1,1);
  interval Z2 = get_j_coef(ITZ2,0);
  interval DZ2 = get_j_coef(ITZ2,1);
  
  // Now, Z1, Z2, DZ1 and DZ2 are intervals
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  itaylor ITZ20 = z2(x_it0,p);
  interval DZ20 = get_j_derive(ITZ20,1);
  
  interval den = cosh(Z2)-cos(Z1);
  if (0 <= den){
    // Division by zero
    flag |= 1;
    return interval(0.0);
  }
  return  (interval)(-2.0)*(DZ10-DZ1)*sin(Z1)*(sinh(Z2)*(DZ20-DZ2) + sin(Z1)*(DZ10-DZ1))/sqr(den);
}



// Integrand of A_{1} (singular part - numerator): itaylor data type
itaylor integrand_A1_singularity_num_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order+2,x);
  itaylor Z1 = z1(x_it,p);
  itaylor D2Z1 = diff(Z1, 2);
  Z1 = trunc(Z1,order);
  
  return  (interval)2.0*D2Z1*sin(Z1);
}

// Integrand of A_{1} (singular part - denominator): itaylor data type
itaylor integrand_A1_singularity_den_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order,x);
  itaylor Z1 = z1(x_it,p);
  itaylor Z2 = z2(x_it,p);
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor den = cosh(Z2)-cos(Z1);
  return den;
}

// Integrand of A_{2} (singular part - numerator): itaylor data type
itaylor integrand_A2_singularity_num_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order+1,x);
  itaylor Z1 = z1(x_it,p);
  itaylor DZ1 = diff(Z1, 1);
  Z1 = trunc(Z1,order);
  
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  
  return  (interval)2.0*sqr(DZ10-DZ1)*cos(Z1);
}

// Integrand of A_{2} (singular part - denominator): itaylor data type
itaylor integrand_A2_singularity_den_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order,x);
  itaylor Z1 = z1(x_it,p);
  itaylor Z2 = z2(x_it,p);
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor den = cosh(Z2)-cos(Z1);
  return den;
}


// Integrand of A_{3} (singular part - numerator): itaylor data type
itaylor integrand_A3_singularity_num_slow(const interval& x, const ParameterSet& p, int order, int& flag){

  itaylor x_it = itaylor(order+1,x);
  itaylor Z1 = z1(x_it,p);
  itaylor Z2 = z2(x_it,p);
  itaylor DZ1 = diff(Z1, 1);
  itaylor DZ2 = diff(Z2, 1);
  Z1 = trunc(Z1,order);
  Z2 = trunc(Z2,order);
  
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor x_it0 = itaylor(1,(interval)0.0);
  itaylor ITZ10 = z1(x_it0,p);
  interval DZ10 = get_j_derive(ITZ10,1);
  itaylor ITZ20 = z2(x_it0,p);
  interval DZ20 = get_j_derive(ITZ20,1);
  
  return  (interval)(-2.0)*(DZ10-DZ1)*sin(Z1)*(sinh(Z2)*(DZ20-DZ2) + sin(Z1)*(DZ10-DZ1));

}

// Integrand of A_{3} (singular part - denominator): itaylor data type
itaylor integrand_A3_singularity_den_slow(const interval& x, const ParameterSet& p, int order, int& flag){
  itaylor x_it = itaylor(order,x);
  itaylor Z1 = z1(x_it,p);
  itaylor Z2 = z2(x_it,p);
  // Now, Z1 and DZ1 are 'order'-order itaylor objects
  itaylor den = cosh(Z2)-cos(Z1);
  return sqr(den);
}

