#include "methods.h"

// Computes a Gauss-Legendre quadrature of order 2 between left_abscissa and right_abscissa.
// To optimize the code, integrand_1D_fast is only interval-based, while
// integrand_1D_slow is itaylor-based.
// Flag returns 1 if division by zero is encountered.
interval GaussLegendre2_1D(const interval& left_abscissa, const interval& right_abscissa,
                              const ParameterSet &p,
                              interval (*integrand_1D_fast)(const interval&, const ParameterSet&, int&),
                              itaylor (*integrand_1D_slow)(const interval&, const ParameterSet&, int, int&),
                              int& flag){
  interval mid_interval = (left_abscissa+right_abscissa)/(interval)2.0;
  interval alpha = left_abscissa|right_abscissa;
  interval dx = right_abscissa - left_abscissa;
  
  interval left = integrand_1D_fast(-dx*sqrt((interval)3.0)/(interval)6.0+(left_abscissa+right_abscissa)/(interval)2.0,p,flag);
  interval right = integrand_1D_fast(dx*sqrt((interval)3.0)/(interval)6.0+(left_abscissa+right_abscissa)/(interval)2.0,p,flag);
  
  itaylor error = integrand_1D_slow(alpha,p,4,flag);
  if (flag & 1){
    return interval(0.0);
  }
  else {
    interval err = get_j_coef(error,4)/(interval)180*sqr(sqr(dx)); // 1/24*error
    return (left + right)/(interval)2.0 + err;  
  }
}


// Computes the singularity part of a given 1D integrand in the form of a fraction.
// order_num and order_den are the orders up to which the numerator and the
// denominator are zero respectively.
// We try to expand in Taylor series around zero and cancel out the powers.
// If this fails (division by zero is encountered), 
// we return a 1D-Gauss-Legendre quadrature of order 2.
// Note that flag will not be set to 1 since the denominator will not have division
// by zero problems, but the fraction num/den will!
interval integrate_singularity_1D(const interval& left_abscissa, const interval& right_abscissa,
                              const ParameterSet &p,
                              itaylor (*integrand_num)(const interval&, const ParameterSet&, int, int&),
                              itaylor (*integrand_den)(const interval&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                              itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&),
                              int& flag, int order_num, int order_den){
  
  interval singul = 0.0 | right_abscissa;
  interval dx = right_abscissa - left_abscissa;
  
  itaylor nums = integrand_num(singul,p,order_num,flag);
  itaylor dens = integrand_den(singul,p,order_den,flag);
  
  // If the Taylor expansion fails, then try Gauss-Legendre.
  if (0 <= get_j_coef(dens,order_den)){
    return GaussLegendre2_1D(left_abscissa,right_abscissa,p,integrand_fast,integrand_slow,flag)*dx;
  }
  interval res = get_j_coef(nums,order_num)/get_j_coef(dens,order_den)
                 /(interval)(order_num-order_den + 1)
                 *(power(right_abscissa,order_num-order_den+1)-power(left_abscissa,order_num-order_den+1));
  return res;
}




// Caller for recursive integration with arithmetic mean splitting.
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_1D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_1D(const ParameterSet& p, const interval& left, const interval& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const interval&, const ParameterSet&, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                          itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level){
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  interval vol = right - left; 
  int curr_flag = 0;
  interval res = integration_method(left,right,p,integrand_fast,integrand_slow,curr_flag)*vol;
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_1D
    if (recursion_level > MAX_DEPTH_RECURSION_1D){
      recursion_level--;
      // Integral calculated although tolerances not met.
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    interval midpoint = (right+left)/(interval)2.0;
    
    recursion_level++;
    new_res+=rec_integrate_1D(p,left,midpoint,integration_method,integrand_fast,integrand_slow,flag,recursion_level);
    
    if (flag & 1){ 
      recursion_level--;
      return (interval)(0.0);
    }
    
    recursion_level++;
    new_res+=rec_integrate_1D(p,midpoint,right,integration_method,integrand_fast,integrand_slow,flag,recursion_level);
    
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}



// Caller for recursive integration with arithmetic mean splitting.
// One has to be careful since the quadratures return the total integral
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_1D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure. 
interval rec_integrate_singularity_1D(const ParameterSet& p, const interval& left, const interval& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         interval (*)(const interval&, const ParameterSet&, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         int&, int, int),
                          itaylor (*integrand_num)(const interval&, const ParameterSet&, int, int&),
                          itaylor (*integrand_den)(const interval&, const ParameterSet&, int, int&), 
                          interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                          itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&),
                          int& flag, int& recursion_level, int order_num, int order_den){
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  interval vol = right - left; 
  int curr_flag = 0;
  interval res = integration_method(left,right,p,integrand_num,integrand_den,integrand_fast,integrand_slow,curr_flag, order_num, order_den);
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_1D_SINGULARITY
    if (recursion_level > MAX_DEPTH_RECURSION_1D_SINGULARITY){
      recursion_level--;
      // Integral calculated although tolerances not met.
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    
    interval midpoint = (right+left)/(interval)2.0;
    recursion_level++;
    new_res+=rec_integrate_singularity_1D(p,left,midpoint,integration_method,
                                       integrand_num,integrand_den,
                                       integrand_fast,integrand_slow,
                                       flag,recursion_level,order_num,order_den);
    if (flag & 1){ 
      recursion_level--;
      return (interval)(0.0);
    }    
    recursion_level++;
    new_res+=rec_integrate_singularity_1D(p,midpoint,right,integration_method,
                                       integrand_num,integrand_den,
                                       integrand_fast,integrand_slow,
                                       flag,recursion_level,order_num,order_den);
                                       
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}

