// The file is split into two types of functions: the quadratures (Newton-Cotes,
// Gauss-Legendre, 1D-2D, etc) and the callers (iterative, recursive, arithmetic
// splitting, geometric splitting, 1D-2D,...). 
// The callers are at an outer level and call the quadratures. With the result,
// depending on the caller it can decide whether to keep it, split the interval
// in different ways and call recursively the quadratures, flag an error, etc.
// The implementation is based on pointers to functions so that changing one
// quadrature to another is easy, same for the integrands, which are called by
// the quadratures. This is intended in order to make experiments with 
// algebraically equal expressions but interval-arithmetic different ones (such
// as for example 1-x^2 and (1-x)*(1+x))

#ifndef METHODS_H_INCLUDED
#define METHODS_H_INCLUDED

#include "curves.h"
#include "interval.hpp"
#include "itaylor.hpp"
#include "dim2taylor.hpp"


// Maximum recursion level for 1D integrals
#define MAX_DEPTH_RECURSION_1D 18
// Maximum recursion level for 1D (singularity) integrals
#define MAX_DEPTH_RECURSION_1D_SINGULARITY 12


// 1D quadratures

// Computes a Gauss-Legendre quadrature of order 2 between left_abscissa and right_abscissa.
interval GaussLegendre2_1D(const interval& left_abscissa, const interval& right_abscissa,
                              const ParameterSet &p,
                              interval (*integrand_1D_fast)(const interval&, const ParameterSet&, int&),
                              itaylor (*integrand_1D_slow)(const interval&, const ParameterSet&, int, int&),
                              int& flag);
                              

// Computes the singularity part of a given 1D integrand in the form of a fraction.
// order_num and order_den are the orders up to which the numerator and the
// denominator are zero respectively.
// If the Taylor expansion fails, the function resorts to Gauss-Legendre of order 2.
interval integrate_singularity_1D(const interval& left_abscissa, const interval& right_abscissa,
                              const ParameterSet &p,
                              itaylor (*integrand_num)(const interval&, const ParameterSet&, int, int&),
                              itaylor (*integrand_den)(const interval&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                              itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&),
                              int& flag, int order_num, int order_den);

// 1D callers

// Caller for recursive integration with arithmetic mean splitting.
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_1D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_1D(const ParameterSet& p, const interval& left, const interval& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const interval&, const ParameterSet&, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                          itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level);
                          
// Caller for recursive integration with arithmetic mean splitting.
// One has to be careful since the quadratures return the total integral
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_1D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.                          
interval rec_integrate_singularity_1D(const ParameterSet& p, const interval& left, const interval& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         interval (*)(const interval&, const ParameterSet&, int&),
                                                         itaylor (*)(const interval&, const ParameterSet&, int, int&),
                                                         int&, int, int),
                          itaylor (*integrand_num)(const interval&, const ParameterSet&, int, int&),
                          itaylor (*integrand_den)(const interval&, const ParameterSet&, int, int&), 
                          interval (*integrand_fast)(const interval&, const ParameterSet&, int&),
                          itaylor (*integrand_slow)(const interval&, const ParameterSet&, int, int&),
                          int& flag, int& recursion_level, int order_num, int order_den);
                          

#endif

