#include "methods.h"


// Computes a Gauss-Legendre quadrature of order 2 in the box
// [y_left,y_right] \times [\gamma_down,\gamma_up]
// To optimize the code, integrand_2D_fast is only interval-based, while
// integrand_2D_slow is dim2taylor-based.
// The function sets flag = 1 if the method fails.
interval GaussLegendre2_2D(const interval& y_left, const interval& y_right,
                              const interval& gamma_down, const interval& gamma_up,
                              const ParameterSet &p,
                              interval (*integrand_2D_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_2D_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag){

  interval dy = y_right - y_left;
  interval dgamma = gamma_up - gamma_down;
  
  // Error
  ivector error_i(2);
  error_i[1] = y_left|y_right; error_i[2] = gamma_up|gamma_down;
  dim2taylor error = integrand_2D_slow(error_i,p,4,flag);
  
  if (flag == 1){
    return interval(0.0);
  }

  // Left Up
  ivector left_up_i(2); 
  left_up_i[1] = -dy/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(y_left+y_right)/(interval)2.0;
  left_up_i[2] =  dgamma/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(gamma_up+gamma_down)/(interval)2.0;
  interval left_up = integrand_2D_fast(left_up_i,p,flag);

  if (flag == 1){
    return interval(0.0);
  }

  // Right Up
  ivector right_up_i(2); 
  right_up_i[1] = dy/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(y_left+y_right)/(interval)2.0;
  right_up_i[2] = dgamma/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(gamma_up+gamma_down)/(interval)2.0;
  interval right_up = integrand_2D_fast(right_up_i,p,flag);
  
  if (flag == 1){
    return interval(0.0);
  }

  // Left Down
  ivector left_down_i(2); 
  left_down_i[1] = -dy/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(y_left+y_right)/(interval)2.0;
  left_down_i[2] = -dgamma/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(gamma_up+gamma_down)/(interval)2.0;
  interval left_down = integrand_2D_fast(left_down_i,p,flag);
  
  if (flag == 1){
    return interval(0.0);
  }

  // Right Down
  ivector right_down_i(2); 
  right_down_i[1] = dy/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(y_left+y_right)/(interval)2.0;
  right_down_i[2] = -dgamma/(interval)2.0*sqrt((interval)3.0)/(interval)3.0+(gamma_up+gamma_down)/(interval)2.0;
  interval right_down = integrand_2D_fast(right_down_i,p,flag);
  
  if (flag == 1){
    return interval(0.0);
  }

  
  interval error_4y = error[4][0];
  interval error_4gamma = error[0][4];

  interval err = error_4y*sqr(sqr(dy)) + error_4gamma*sqr(sqr(dgamma));    
  
  
  return (left_up+right_up+left_down+right_down)/(interval)4.0
          +err/(interval)180.0;
}



// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_x, order_num_y, order_den_x and order_den_y are the orders up to which the numerator and the
// denominator are zero respectively in x and y.
// We assume that x_left*x_right >= 0 and that y_down*y_up >= 0.
// We do the Taylor expansion in both variables and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.
interval integrate_singularity_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y){

  interval dx = x_right - x_left;
  interval dy = y_up - y_down;
  ivector error_i(2);
  error_i[1] = (0.0|x_left)|x_right; error_i[2] = (0.0|y_down)|y_up;
  
  dim2taylor nums = integrand_num_slow(error_i,p,order_num_x+order_num_y,flag);
  dim2taylor dens = integrand_den_slow(error_i,p,order_den_x+order_den_y,flag);
  
  interval vol = dx*dy;
  
  
  // If the Taylor expansion fails, then return Gauss-Legendre.
  if (0 <= dens[order_den_x][order_den_y]){
    return GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,flag)*vol;
  }
  
  // Otherwise return the intersection of the two
  interval res1 = nums[order_num_x][order_num_y]/dens[order_den_x][order_den_y]
                 /(interval)(order_num_x-order_den_x+1)
                 /(interval)(order_num_y-order_den_y+1)
                 *(power(x_right,order_num_x-order_den_x+1)-power(x_left,order_num_x-order_den_x+1))
                 *(power(y_up,order_num_y-order_den_y+1)-power(y_down,order_num_y-order_den_y+1));
  int curr_flag = 0;
  interval res2 = GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,curr_flag)*vol;
  
  if (curr_flag & 1){
    return res1;
  }
  else {
    return res1 & res2;
  }
                              
}


// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_x and order_den_x are the orders up to which the numerator and the
// denominator are zero respectively in x.
// order_num_y and order_den_y are ignored and only used for compatibility with the caller.
// We assume that x_left*x_right >= 0.
// We do the Taylor expansion in the x variable and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.
interval integrate_singularity_x_axis_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y){

  interval dx = x_right - x_left;
  interval dy = y_up - y_down;
  ivector error_i(2);
  error_i[1] = (0.0|x_left)|x_right; error_i[2] = y_down|y_up;
  
  dim2taylor nums = integrand_num_slow(error_i,p,order_num_x,flag);
  dim2taylor dens = integrand_den_slow(error_i,p,order_den_x,flag);
  
  interval vol = dx*dy;
  
  
  // If the Taylor expansion fails, then return Gauss-Legendre.
  if (0 <= dens[order_den_x][0]){
    return GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,flag)*vol;
  }
  
  // Otherwise return the intersection of the two
  interval res1 = nums[order_num_x][0]/dens[order_den_x][0]
                 /(interval)(order_num_x-order_den_x+1)
                 *(power(x_right,order_num_x-order_den_x+1)-power(x_left,order_num_x-order_den_x+1))*dy;
  int curr_flag = 0;
  interval res2 = GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,curr_flag)*vol;
  
  if (curr_flag & 1){
    return res1;
  }
  else {
    return res1 & res2;
  }
                              
}

// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_y and order_den_y are the orders up to which the numerator and the
// denominator are zero respectively in y.
// order_num_x and order_den_x are ignored and only used for compatibility with the caller.
// We assume that y_down*y_up >= 0.
// We do the Taylor expansion in the y variable and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.
interval integrate_singularity_y_axis_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y){

  interval dx = x_right - x_left;
  interval dy = y_up - y_down;
  ivector error_i(2);
  error_i[1] = x_left|x_right; error_i[2] = (0.0|y_down)|y_up;
  
  dim2taylor nums = integrand_num_slow(error_i,p,order_num_y,flag);
  dim2taylor dens = integrand_den_slow(error_i,p,order_den_y,flag);
  
  interval vol = dx*dy;
      
  // If the Taylor expansion fails, then return Gauss-Legendre.
  if (0 <= dens[0][order_den_y]){
    return GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,flag)*vol;
  }
  
  // Otherwise return the intersection of the two
  interval res1 = nums[0][order_num_y]/dens[0][order_den_y]
                 /(interval)(order_num_y-order_den_y+1)
                 *(power(y_up,order_num_y-order_den_y+1)-power(y_down,order_num_y-order_den_y+1))*dx;
  int curr_flag = 0;
  interval res2 = GaussLegendre2_2D(x_left,x_right,y_down,y_up,p,integrand_fast,integrand_slow,curr_flag)*vol;
  
  if (curr_flag & 1){
    return res1;
  }
  else {
    return res1 & res2;
  }
                              
}


// 2D callers

// Caller for recursive integration with arithmetic mean splitting.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_2D(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level){
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  ivector measure; measure = (right-left);
  interval vol(1.0); 
  for (int i = 1; i <= VecLen(measure); i++){
    vol*= measure[i];
  }
  int curr_flag = 0;
  interval res = integration_method(left[1],right[1],left[2],right[2],p,integrand_fast,integrand_slow,curr_flag)*vol;
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_2D
    if (recursion_level > MAX_DEPTH_RECURSION_2D){
      recursion_level--;
      // Integral calculated although tolerances not met.
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    ivector midpoint = (right+left)/2.0;
    
    for (int i = 0; i < (1 << VecLen(left)); i++){
      ivector curr_left;
      ivector curr_right;
      curr_left = midpoint; curr_right = midpoint;
      for (int j = 0; j < VecLen(left); j++){
        if (i & (1 << j)){
          curr_left[j+1] = left[j+1];
        }
        else {
          curr_right[j+1] = right[j+1];
        }
      } 
      recursion_level++;
      new_res+=rec_integrate_2D(p,curr_left,curr_right,integration_method,integrand_fast,integrand_slow,flag,recursion_level);  

      if (flag & 1){ 
        recursion_level--;
        return (interval)(0.0);
      }
    }
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}



// Caller for recursive integration with geometric mean splitting.
// Only the first 'num_coords' coordinates are split geometrically, the rest
// are split arithmetically.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_2D_geometric(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int num_coords){        
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  ivector measure; measure = (right-left);
  interval vol(1.0); 
  for (int i = 1; i <= VecLen(measure); i++){
    vol*= measure[i];
  }
  int curr_flag = 0;
  interval res = integration_method(left[1],right[1],left[2],right[2],p,integrand_fast,integrand_slow,curr_flag)*vol;
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_2D
    if (recursion_level > MAX_DEPTH_RECURSION_2D){
      // Integral calculated although tolerances not met.
      recursion_level--;
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    
    ivector midpoint = (right+left)/2.0;
    ivector gmean(VecLen(left));
    for (int i = 1; i <= VecLen(gmean); i++){
      gmean[i] = sqrt(left[i]*right[i]);
      if (Sup(left[i]) < (double)0) gmean[i] = -gmean[i];
    }
    for (int i = 0; i < (1 << VecLen(left)); i++){
      ivector curr_left;
      ivector curr_right;
      curr_left = midpoint; curr_right = midpoint;
      for (int j = 0; j < VecLen(left); j++){
        // First num_coords coordinates
        if (j < num_coords && (i & (1 << j))){
          curr_left[j+1] = left[j+1];
          curr_right[j+1] = gmean[j+1];
        }
        else if (j < num_coords && (i & (1 << j)) != (1 << j)){
          curr_left[j+1] = gmean[j+1];
          curr_right[j+1] = right[j+1];
        }
        // Last n - num_coords coordinates
        else if (j >= num_coords && (i & (1 << j))){
          curr_left[j+1] = left[j+1];
          curr_right[j+1] = midpoint[j+1];
        }
        else if (j >= num_coords && (i & (1 << j)) != (1 << j)){
          curr_left[j+1] = midpoint[j+1];
          curr_right[j+1] = right[j+1];
        }
      } 
      recursion_level++;
      new_res+=rec_integrate_2D_geometric(p,curr_left,curr_right,integration_method,
                                          integrand_fast,integrand_slow,
                                          flag,recursion_level,num_coords);  

      if (flag & 1){ 
        recursion_level--;
        return (interval)(0.0);
      }
    }
    
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}

// Caller for recursive integration with arithmetic mean splitting.
// Intended for singularities in the form of a fraction of orders (num_x, num_y) and (den_x, den_y)
// in the numerator and denominator respectively.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral 
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure. 
interval rec_integrate_singularity_2D(const ParameterSet& p, const ivector& left, const ivector& right,
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&, int, int, int, int),
                          dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                          dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int order_num_x, int order_num_y, int order_den_x, int order_den_y){
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  ivector measure; measure = (right-left);
  interval vol(1.0); 
  for (int i = 1; i <= VecLen(measure); i++){
    vol*= measure[i];
  }
  int curr_flag = 0;
  interval res = integration_method(left[1],right[1],left[2],right[2],p,
                                    integrand_num_slow,integrand_den_slow,
                                    integrand_fast,integrand_slow,
                                    curr_flag,order_num_x,order_num_y,
                                    order_den_x,order_den_y);
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_2D_SINGULARITY
    if (recursion_level > MAX_DEPTH_RECURSION_2D_SINGULARITY){
      // Integral calculated although tolerances not met.
      recursion_level--;
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    
    ivector midpoint = (right+left)/2.0;
    for (int i = 0; i < (1 << VecLen(left)); i++){
      ivector curr_left;
      ivector curr_right;
      curr_left = midpoint; curr_right = midpoint;
      for (int j = 0; j < VecLen(left); j++){
        if (i & (1 << j)){
          curr_left[j+1] = left[j+1];
        }
        else {
          curr_right[j+1] = right[j+1];
        }
      } 
      recursion_level++;
      new_res+=rec_integrate_singularity_2D(p,curr_left,curr_right,integration_method,
                                            integrand_num_slow,integrand_den_slow,
                                            integrand_fast,integrand_slow,
                                            flag,recursion_level,order_num_x,order_num_y,
                                            order_den_x,order_den_y);  

      if (flag & 1){ 
        recursion_level--;
        return (interval)(0.0);
      }
    }
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}


// Caller for recursive integration with geometric mean splitting.
// Only the first 'num_coords' coordinates are split geometrically, the rest
// are split arithmetically.
// Intended for singularities in the form of a fraction of orders (num_x, num_y) and (den_x, den_y)
// in the numerator and denominator respectively.
// The geometric splitting may be done in the non-singular coordinates, since otherwise it might
// result in [0,0] intervals if 0 belongs to that coordinate.
// The singular coordinates should be split arithmetically.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral 
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure. 
interval rec_integrate_singularity_2D_geometric(const ParameterSet& p, const ivector& left, const ivector& right,
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&, int, int, int, int),
                          dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                          dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int order_num_x, int order_num_y, int order_den_x, int order_den_y,
                          int num_coords){
  if (flag & 1){ 
    recursion_level--;
    return (interval)(0.0);
  }
  ivector measure; measure = (right-left);
  interval vol(1.0); 
  for (int i = 1; i <= VecLen(measure); i++){
    vol*= measure[i];
  }
  int curr_flag = 0;
  interval res = integration_method(left[1],right[1],left[2],right[2],p,
                                    integrand_num_slow,integrand_den_slow,
                                    integrand_fast,integrand_slow,
                                    curr_flag,order_num_x,order_num_y,
                                    order_den_x,order_den_y);
  // We bisect either if we don't meet the tolerances or if we divide by zero
  if (diam(res) > p.abs_tol || diam(res)/mid(vol) > p.rel_tol || curr_flag != 0){
    // Depth beyond MAX_DEPTH_RECURSION_2D_SINGULARITY
    if (recursion_level > MAX_DEPTH_RECURSION_2D_SINGULARITY){
      // Integral calculated although tolerances not met.
      recursion_level--;
      flag |= 2;
      if (curr_flag & 1) flag |= 1;
      return res;
    }
    // bisection
    interval new_res(0.0);
    
    ivector midpoint = (right+left)/2.0;
    ivector gmean(VecLen(left));
    for (int i = 1; i <= VecLen(gmean); i++){
      gmean[i] = sqrt(left[i]*right[i]);
      if (Sup(left[i]) < (double)0) gmean[i] = -gmean[i];
    }
    for (int i = 0; i < (1 << VecLen(left)); i++){
      ivector curr_left;
      ivector curr_right;
      curr_left = midpoint; curr_right = midpoint;
      for (int j = 0; j < VecLen(left); j++){
        // First num_coords coordinates
        if (j < num_coords && (i & (1 << j))){
          curr_left[j+1] = left[j+1];
          curr_right[j+1] = gmean[j+1];
        }
        else if (j < num_coords && (i & (1 << j)) != (1 << j)){
          curr_left[j+1] = gmean[j+1];
          curr_right[j+1] = right[j+1];
        }
        // Last n - num_coords coordinates
        else if (j >= num_coords && (i & (1 << j))){
          curr_left[j+1] = left[j+1];
          curr_right[j+1] = midpoint[j+1];
        }
        else if (j >= num_coords && (i & (1 << j)) != (1 << j)){
          curr_left[j+1] = midpoint[j+1];
          curr_right[j+1] = right[j+1];
        }
      } 
      recursion_level++;
      new_res+=rec_integrate_singularity_2D_geometric(p,curr_left,curr_right,integration_method,
                                            integrand_num_slow,integrand_den_slow,
                                            integrand_fast,integrand_slow,
                                            flag,recursion_level,order_num_x,order_num_y,
                                            order_den_x,order_den_y,num_coords);  

      if (flag & 1){ 
        recursion_level--;
        return (interval)(0.0);
      }
    }
    recursion_level--;
    return new_res;
  }
  else {
    recursion_level--;
    return res;
  }
}

