// The file is split into two types of functions: the quadratures (Newton-Cotes,
// Gauss-Legendre, 1D-2D, etc) and the callers (iterative, recursive, arithmetic
// splitting, geometric splitting, 1D-2D,...). 
// The callers are at an outer level and call the quadratures. With the result,
// depending on the caller it can decide whether to keep it, split the interval
// in different ways and call recursively the quadratures, flag an error, etc.
// The implementation is based on pointers to functions so that changing one
// quadrature to another is easy, same for the integrands, which are called by
// the quadratures. This is intended in order to make experiments with 
// algebraically equal expressions but interval-arithmetic different ones (such
// as for example 1-x^2 and (1-x)*(1+x))

#ifndef METHODS_H_INCLUDED
#define METHODS_H_INCLUDED

#include "curves.h"
#include "interval.hpp"
#include "itaylor.hpp"
#include "dim2taylor.hpp"


// Maximum recursion level for 1D integrals
#define MAX_DEPTH_RECURSION_1D 18
// Maximum recursion level for 1D (singularity) integrals
#define MAX_DEPTH_RECURSION_1D_SINGULARITY 12

#define MAX_DEPTH_RECURSION_2D 10

#ifdef CASE_SINGULAR_CENTER
  #define MAX_DEPTH_RECURSION_2D_SINGULARITY 8
#elif CASE_SINGULAR_FIRST
  #define MAX_DEPTH_RECURSION_2D_SINGULARITY 8
#elif CASE_SINGULAR_SECOND
  #define MAX_DEPTH_RECURSION_2D_SINGULARITY 9
#elif CASE_SINGULAR_SECOND_SPECIAL
  #define MAX_DEPTH_RECURSION_2D_SINGULARITY 10
#elif CASE_NONSINGULAR
  #define MAX_DEPTH_RECURSION_2D_SINGULARITY 8
#else
  #error Wrong values of CASE, constant not set by preprocessor
#endif

// 2D quadratures

// Computes a Gauss-Legendre quadrature of order 2 in the box
// [y_left,y_right] \times [\gamma_down,\gamma_up]
interval GaussLegendre2_2D(const interval& y_left, const interval& y_right,
                              const interval& gamma_down, const interval& gamma_up,
                              const ParameterSet &p,
                              interval (*integrand_2D_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_2D_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag);



// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_x, order_num_y, order_den_x and order_den_y are the orders up to which the numerator and the
// denominator are zero respectively in x and y.
// We assume that x_left*x_right >= 0 and that y_down*y_up >= 0.
// We do the Taylor expansion in both variables and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.
interval integrate_singularity_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y);


// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_x and order_den_x are the orders up to which the numerator and the
// denominator are zero respectively in x.
// order_num_y and order_den_y are ignored and only used for compatibility with the caller.
// We assume that x_left*x_right >= 0.
// We do the Taylor expansion in the x variable and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.
interval integrate_singularity_x_axis_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y);

                              
// Computes the singularity part of a given 2D integrand in the form of a fraction.
// order_num_y and order_den_y are the orders up to which the numerator and the
// denominator are zero respectively in y.
// order_num_x and order_den_x are ignored and only used for compatibility with the caller.
// We assume that y_down*y_up >= 0.
// We do the Taylor expansion in the y variable and integrate.
// If the Taylor expansion fails, the function resorts to a two-dimensional 
// Gauss-Legendre quadrature of order 2.
// Otherwise, it returns the intersection of both approaches.                             
interval integrate_singularity_y_axis_2D(const interval& x_left, const interval& x_right,
                              const interval& y_down, const interval& y_up,
                              const ParameterSet &p,
                              dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                              dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                              interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                              dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&),
                              int& flag, int order_num_x, int order_num_y, int order_den_x, int order_den_y);


// 2D callers                          

// Caller for recursive integration with arithmetic mean splitting.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_2D(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level);



// Caller for recursive integration with geometric mean splitting.
// Only the first 'num_coords' coordinates are split geometrically, the rest
// are split arithmetically.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral divided
// by the volume.
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.
interval rec_integrate_2D_geometric(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int num_coords);
                          

// Caller for recursive integration with arithmetic mean splitting.
// Intended for singularities in the form of a fraction of orders (num_x, num_y) and (den_x, den_y)
// in the numerator and denominator respectively.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral 
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.                          
interval rec_integrate_singularity_2D(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&, int, int, int, int),
                          dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                          dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int order_num_x, int order_num_y, int order_den_x, int order_den_y);



// Caller for recursive integration with geometric mean splitting.
// Only the first 'num_coords' coordinates are split geometrically, the rest
// are split arithmetically.
// Intended for singularities in the form of a fraction of orders (num_x, num_y) and (den_x, den_y)
// in the numerator and denominator respectively.
// The geometric splitting may be done in the non-singular coordinates, since otherwise it might
// result in [0,0] intervals if 0 belongs to that coordinate.
// The singular coordinates should be split arithmetically.
// Left and Right are vectors with the lower (resp. upper) bounds in each of the directions. 
// One has to be careful since the quadratures return the total integral 
// (not the one divided by the volume, as opposed to other callers).
// The number of recursive subdivisions is controlled by MAX_DEPTH_RECURSION_2D_SINGULARITY.
// If we encounter a division by zero at the deepest level, flag will have
// its last bit set to one (critical failure).
// If we are not able to meet the tolerance requirements (but still able to
// compute an enclosure of the true value), flag will have its second last
// bit set to one (as a warning) and will return the enclosure.                          
interval rec_integrate_singularity_2D_geometric(const ParameterSet& p, const ivector& left, const ivector& right, 
                          interval (*integration_method)(const interval&, const interval&,
                                                         const interval&, const interval&,
                                                         const ParameterSet&, 
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         interval (*)(const ivector&, const ParameterSet&, int&),
                                                         dim2taylor (*)(const ivector&, const ParameterSet&, int, int&),
                                                         int&, int, int, int, int),
                          dim2taylor (*integrand_num_slow)(const ivector&, const ParameterSet&, int, int&),
                          dim2taylor (*integrand_den_slow)(const ivector&, const ParameterSet&, int, int&),
                          interval (*integrand_fast)(const ivector&, const ParameterSet&, int&),
                          dim2taylor (*integrand_slow)(const ivector&, const ParameterSet&, int, int&), 
                          int& flag, int& recursion_level, int order_num_x, int order_num_y, int order_den_x, int order_den_y,
                          int num_coords);                          
                          

#endif

