// Computation of the second part of Lemma 2.2.

// The abstract setting is the following: we will have 'callers' which will call
// 'quadratures', which will call 'integrands'. Different callers decide for example
// when and how to split the integration domains in a different way, the quadratures
// implement different integration methods and the integrands are the different
// functions under the integral sign.


// We validate \pa_{tt} \pa_x z^{1}(0,0) for A(\eps) = [1.08050,1.08055], \eps = [0,1e-6],
// showing a uniform positive lower bound, independent of \eps.

// The previous quantity consists of 41 2D integrals: each of them we split it
// into singular and non-singular regions. The singular regions are either
// a neighbourhood of the origin, a neighbourhood of the first-coordinate axis,
// or a neighbourhood of the second-coordinate axis. The notation follows 
// closely the paper.

// The singular regions are calculated expanding both the numerator
// and the denominator in Taylor series to avoid fractions of the type 0/0.
// The non-singular ones are calculated using a Gauss-Legendre quadrature of
// order 2. All integrations are done adaptively, bisecting the region of
// integration whenever some tolerances are not met, up to a maximum number of
// bisections. The bisection in each coordinate can be done either by the midpoint
// (arithmetic mean of the endpoints) or by the geometric mean of the midpoints,
// depending on the type of integral we are calculating.

// The program is split in 3(x2) more files:

// curves.h/cpp contains information about the family of curves and its
// implementation with the different data structures (interval,itaylor,dim2taylor).
// It also contains the methods and the definition of the class ParameterSet,
// which will enclose all the information we need to perform the calculations.

// integrands.h/cpp contains the different integrands of the functions to be computed.

// methods.h/cpp contains both the caller and the quadrature methods.

// All calculations are done using rigorous arithmetics using the C-XSC library.

// Usage: ./Lemma_222_X(_Y)
// X = {"nonsingular","singular_center","singular_first_coord",
//      "singular_second_coord","singular_second_coord_special"}
// depending on the region and the integral to compute.
// Y = {1,2,3,4,5,6,7} in the "singular_second_coord" case.
// Y = {1,2,3,4,5,6} in the "singular_second_coord_special" case.

// The output is to the screen.
// For reference, the output for the different cases is located at /input_output.



#include "interval.hpp"
#include "itaylor.hpp"
#include "dim2taylor.hpp"
#include "curves.h"
#include "integrands.h"
#include "methods.h"

#include <iostream>
#include <string>
#include <cassert>
#include <vector>

using namespace cxsc;
using namespace std;
using namespace taylor;



///////////////////////////////////////////////////////////
// BEGIN DEFINITION OF THE INTEGRALS TO BE COMPUTED
///////////////////////////////////////////////////////////


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [\delta,\pi] 
// \cup [\delta,\pi] \times [-\pi,-\delta] (nonsingular region). The splitting
// is done in a geometric way.


// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_nonsingular(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = p.Left;
  ivector right1(2); right1[1] = p.Right; right1[2] = p.Right;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Right;
  ivector right2(2); right2[1] = p.Right; right2[2] = -p.Left;
  
  interval res2d_nonsing(0.0);
  interval curr_res;
  
  int rec_level = 0; rec_level++; 
  
  // B11
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,2));                              
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B11 " << curr_res << endl; res2d_nonsing += curr_res;
  // B12
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2));                              
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B12 " << curr_res << endl; res2d_nonsing += curr_res;
  // B13
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B13 " << curr_res << endl; res2d_nonsing += curr_res;
  // B14
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B14 " << curr_res << endl; res2d_nonsing += curr_res;
  // B15
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B15 " << curr_res << endl; res2d_nonsing += curr_res;
  // B16
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B16 " << curr_res << endl; res2d_nonsing += curr_res;
  // B21
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B21 " << curr_res << endl; res2d_nonsing += curr_res;
  // B22
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B22 " << curr_res << endl; res2d_nonsing += curr_res;
  // B23
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B23 " << curr_res << endl; res2d_nonsing += curr_res;
  // B24
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B24 " << curr_res << endl; res2d_nonsing += curr_res;
  // B25
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B25 " << curr_res << endl; res2d_nonsing += curr_res;
  // B31
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B31 " << curr_res << endl; res2d_nonsing += curr_res;
  // B32
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B32 " << curr_res << endl; res2d_nonsing += curr_res;
  // B33
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B33 " << curr_res << endl; res2d_nonsing += curr_res;
  // B34
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B34 " << curr_res << endl; res2d_nonsing += curr_res;
  // B35
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Integrand B35 " << curr_res << endl; res2d_nonsing += curr_res;
  // B41
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B41 " << curr_res << endl; res2d_nonsing += curr_res;
  // B42
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B42 " << curr_res << endl; res2d_nonsing += curr_res;
  // B43
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B43 " << curr_res << endl; res2d_nonsing += curr_res;
  // B44
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B44 " << curr_res << endl; res2d_nonsing += curr_res;
  // B45
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B45 " << curr_res << endl; res2d_nonsing += curr_res;
  // B46
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B46 " << curr_res << endl; res2d_nonsing += curr_res;
  // B47
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B47 " << curr_res << endl; res2d_nonsing += curr_res;
  // B51
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Integrand B51 " << curr_res << endl; res2d_nonsing += curr_res;
  // B52
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B52 " << curr_res << endl; res2d_nonsing += curr_res;
  // B53
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B53 " << curr_res << endl; res2d_nonsing += curr_res;
  // B54
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Integrand B54 " << curr_res << endl; res2d_nonsing += curr_res;
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Integrand B55 " << curr_res << endl; res2d_nonsing += curr_res;
  // B61
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B61 " << curr_res << endl; res2d_nonsing += curr_res;
  // B62
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B62 " << curr_res << endl; res2d_nonsing += curr_res;
  // B63
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B63 " << curr_res << endl; res2d_nonsing += curr_res;
  // B64
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B64 " << curr_res << endl; res2d_nonsing += curr_res;
  // B65
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B65 " << curr_res << endl; res2d_nonsing += curr_res;
  // B66
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B66 " << curr_res << endl; res2d_nonsing += curr_res;
  // B67
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B67 " << curr_res << endl; res2d_nonsing += curr_res;
  // B71
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B71 " << curr_res << endl; res2d_nonsing += curr_res;
  // B72
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B72 " << curr_res << endl; res2d_nonsing += curr_res;
  // B73
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B73 " << curr_res << endl; res2d_nonsing += curr_res;
  // B74
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B74 " << curr_res << endl; res2d_nonsing += curr_res;
  // B75
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Integrand B75 " << curr_res << endl; res2d_nonsing += curr_res;
  // B76
  curr_res = (interval)2.0*(
             rec_integrate_2D_geometric(p,left1,right1,GaussLegendre2_2D,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_2D_geometric(p,left2,right2,GaussLegendre2_2D,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,2));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0);
  cerr << "Integrand B76 " << curr_res << endl; res2d_nonsing += curr_res; 
  return res2d_nonsing;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [-\delta,\delta] \times [0,\delta] 
// (singular center). The integration is done by expanding in a Taylor series
// around (0,0) in two variables both numerator and denominator and 
// cancelling the resulting powers. If this fails, then we resort to 
// Gauss-Legendre of order 2. The splitting is done in an arithmetic way.

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_center(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = 0; left1[2] = 0;
  ivector right1(2); right1[1] = p.Left; right1[2] = p.Left;
  ivector left2(2); left2[1] = -p.Left; left2[2] = 0;
  ivector right2(2); right2[1] = 0; right2[2] = p.Left;
  
  interval res2d_sing_center(0.0);
  interval curr_res;
  
  int rec_level = 0; rec_level++; 
  
  // B11
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B11 " << curr_res << endl; res2d_sing_center += curr_res;
  // B12
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B12 " << curr_res << endl; res2d_sing_center += curr_res;
  // B13
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B13 " << curr_res << endl; res2d_sing_center += curr_res;
  // B14
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B14 " << curr_res << endl; res2d_sing_center += curr_res;
  // B15
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B15 " << curr_res << endl; res2d_sing_center += curr_res;
  // B16
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B16 " << curr_res << endl; res2d_sing_center += curr_res;
  // B21
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B21 " << curr_res << endl; res2d_sing_center += curr_res;
  // B22
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B22 " << curr_res << endl; res2d_sing_center += curr_res;
  // B23
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B23 " << curr_res << endl; res2d_sing_center += curr_res;
  // B24
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B24 " << curr_res << endl; res2d_sing_center += curr_res;
  // B25
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B25 " << curr_res << endl; res2d_sing_center += curr_res;
  // B31
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B31 " << curr_res << endl; res2d_sing_center += curr_res;
  // B32
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B32 " << curr_res << endl; res2d_sing_center += curr_res;
  // B33
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B33 " << curr_res << endl; res2d_sing_center += curr_res;
  // B34
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B34 " << curr_res << endl; res2d_sing_center += curr_res;
  // B35
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B35 " << curr_res << endl; res2d_sing_center += curr_res;
  // B41
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Center singularity B41 " << curr_res << endl; res2d_sing_center += curr_res;
  // B42
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B42 " << curr_res << endl; res2d_sing_center += curr_res;
  // B43
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B43 " << curr_res << endl; res2d_sing_center += curr_res;
  // B44
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B44 " << curr_res << endl; res2d_sing_center += curr_res;
  // B45
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B45 " << curr_res << endl; res2d_sing_center += curr_res;
  // B46
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  cerr << "Center singularity B46 " << curr_res << endl; res2d_sing_center += curr_res;
  // B47
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B47 " << curr_res << endl; res2d_sing_center += curr_res;
  // B51
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B51 " << curr_res << endl; res2d_sing_center += curr_res;
  // B52
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B52 " << curr_res << endl; res2d_sing_center += curr_res;
  // B53
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B53 " << curr_res << endl; res2d_sing_center += curr_res;
  // B54
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B54 " << curr_res << endl; res2d_sing_center += curr_res;
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B55 " << curr_res << endl; res2d_sing_center += curr_res;
  // B61
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B61 " << curr_res << endl; res2d_sing_center += curr_res;
  // B62
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B62 " << curr_res << endl; res2d_sing_center += curr_res;
  // B63
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B63 " << curr_res << endl; res2d_sing_center += curr_res;
  // B64
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B64 " << curr_res << endl; res2d_sing_center += curr_res;
  // B65
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B65 " << curr_res << endl; res2d_sing_center += curr_res;
  // B66
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B66 " << curr_res << endl; res2d_sing_center += curr_res;
  // B67
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B67 " << curr_res << endl; res2d_sing_center += curr_res;
  // B71
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B71 " << curr_res << endl; res2d_sing_center += curr_res;
  // B72
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B72 " << curr_res << endl; res2d_sing_center += curr_res;
  // B73
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B73 " << curr_res << endl; res2d_sing_center += curr_res;
  // B74
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B74 " << curr_res << endl; res2d_sing_center += curr_res;
  // B75
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B75 " << curr_res << endl; res2d_sing_center += curr_res;
  // B76
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Center singularity B76 " << curr_res << endl; res2d_sing_center += curr_res;
  return res2d_sing_center;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [-\delta,\delta] \times [\delta,\pi]
// (singular first-coordinate). The integration is done by expanding in a Taylor series
// around (0,z) in the y-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2. The splitting is done in an arithmetic way.

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_first_var(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = 0; left1[2] = p.Left;
  ivector right1(2); right1[1] = p.Left; right1[2] = p.Right;
  ivector left2(2); left2[1] = -p.Left; left2[2] = p.Left;
  ivector right2(2); right2[1] = 0; right2[2] = p.Right;
  
  interval res2d_sing_first_var(0.0);
  interval curr_res;
  
  int rec_level = 0; rec_level++;
  
  // B11
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B11 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B12
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B12 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B13
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B13 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B14
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B14 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B15
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B15 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B16
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B16 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B21
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B21 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B22
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B22 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B23
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B23 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B24
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B24 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B25
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B25 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B31
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B31 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B32
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B32 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B33
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B33 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B34
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B34 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B35
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B35 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B41
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B41 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B42
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B42 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B43
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B43 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B44
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B44 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B45
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B45 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B46
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B46 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B47
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B47 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B51
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B51 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B52
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B52 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B53
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B53 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B54
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B54 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B55 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B61
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B61 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B62
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B62 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B63
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B63 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B64
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B64 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B65
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B65 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B66
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B66 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B67
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B67 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B71
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B71 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B72
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B72 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B73
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B73 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B74
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B74 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B75
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B75 " << curr_res << endl; res2d_sing_first_var += curr_res;
  // B76
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_x_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_x_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Y-axis singularity B76 " << curr_res << endl; res2d_sing_first_var += curr_res;
  return res2d_sing_first_var;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{11} to B_{16} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_1(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);
  
  int rec_level = 0; rec_level++;
  
  // B11
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B11_singularity_num_slow,
                              integrand_B11_singularity_den_slow,
                              integrand_B11_fast,integrand_B11_slow,
                              flag,rec_level,6,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B11 " << curr_res << endl; res2d_sing_second_var += curr_res;

  // B12
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B12_singularity_num_slow,
                              integrand_B12_singularity_den_slow,
                              integrand_B12_fast,integrand_B12_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B12 " << curr_res << endl; res2d_sing_second_var += curr_res;

  // B13
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B13_singularity_num_slow,
                              integrand_B13_singularity_den_slow,
                              integrand_B13_fast,integrand_B13_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B13 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B14
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B14_singularity_num_slow,
                              integrand_B14_singularity_den_slow,
                              integrand_B14_fast,integrand_B14_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B14 " << curr_res << endl; res2d_sing_second_var += curr_res;

  // B15
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B15_singularity_num_slow,
                              integrand_B15_singularity_den_slow,
                              integrand_B15_fast,integrand_B15_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B15 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B16
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B16_singularity_num_slow,
                              integrand_B16_singularity_den_slow,
                              integrand_B16_fast,integrand_B16_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B16 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{21} to B_{25} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.

interval compute_Lemma_222_singular_second_var_split_2(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B21
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B21_singularity_num_slow,
                              integrand_B21_singularity_den_slow,
                              integrand_B21_fast,integrand_B21_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B21 " << curr_res << endl; res2d_sing_second_var += curr_res;

  // B22
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B22_singularity_num_slow,
                              integrand_B22_singularity_den_slow,
                              integrand_B22_fast,integrand_B22_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B22 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B23
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B23_singularity_num_slow,
                              integrand_B23_singularity_den_slow,
                              integrand_B23_fast,integrand_B23_slow,
                              flag,rec_level,2,6,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B23 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B24
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B24_singularity_num_slow,
                              integrand_B24_singularity_den_slow,
                              integrand_B24_fast,integrand_B24_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B24 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B25
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B25_singularity_num_slow,
                              integrand_B25_singularity_den_slow,
                              integrand_B25_fast,integrand_B25_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B25 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}  

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{31} to B_{35} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_3(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
    
  // B31
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B31_singularity_num_slow,
                              integrand_B31_singularity_den_slow,
                              integrand_B31_fast,integrand_B31_slow,
                              flag,rec_level,3,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B31 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B32
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B32_singularity_num_slow,
                              integrand_B32_singularity_den_slow,
                              integrand_B32_fast,integrand_B32_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B32 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B33
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B33_singularity_num_slow,
                              integrand_B33_singularity_den_slow,
                              integrand_B33_fast,integrand_B33_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B33 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B34
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B34_singularity_num_slow,
                              integrand_B34_singularity_den_slow,
                              integrand_B34_fast,integrand_B34_slow,
                              flag,rec_level,2,4,2,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B34 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B35
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B35_singularity_num_slow,
                              integrand_B35_singularity_den_slow,
                              integrand_B35_fast,integrand_B35_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B35 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;

}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{41} to B_{46} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_4(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;  
  
  // B41 
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B41_singularity_num_slow,
                              integrand_B41_singularity_den_slow,
                              integrand_B41_fast,integrand_B41_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B41 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B42
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B42_singularity_num_slow,
                              integrand_B42_singularity_den_slow,
                              integrand_B42_fast,integrand_B42_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B42 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B43
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B43_singularity_num_slow,
                              integrand_B43_singularity_den_slow,
                              integrand_B43_fast,integrand_B43_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B43 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B44
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B44_singularity_num_slow,
                              integrand_B44_singularity_den_slow,
                              integrand_B44_fast,integrand_B44_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B44 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B45
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B45_singularity_num_slow,
                              integrand_B45_singularity_den_slow,
                              integrand_B45_fast,integrand_B45_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B45 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B46
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B46_singularity_num_slow,
                              integrand_B46_singularity_den_slow,
                              integrand_B46_fast,integrand_B46_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B46 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{51} to B_{54} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_5(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B51
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B51_singularity_num_slow,
                              integrand_B51_singularity_den_slow,
                              integrand_B51_fast,integrand_B51_slow,
                              flag,rec_level,3,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B51 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B52
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B52_singularity_num_slow,
                              integrand_B52_singularity_den_slow,
                              integrand_B52_fast,integrand_B52_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B52 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B53
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B53_singularity_num_slow,
                              integrand_B53_singularity_den_slow,
                              integrand_B53_fast,integrand_B53_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B53 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B54
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B54_singularity_num_slow,
                              integrand_B54_singularity_den_slow,
                              integrand_B54_fast,integrand_B54_slow,
                              flag,rec_level,2,8,2,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B54 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{61} to B_{67} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_6(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B61
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B61_singularity_num_slow,
                              integrand_B61_singularity_den_slow,
                              integrand_B61_fast,integrand_B61_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B61 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B62
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B62_singularity_num_slow,
                              integrand_B62_singularity_den_slow,
                              integrand_B62_fast,integrand_B62_slow,
                              flag,rec_level,4,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B62 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B63
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B63_singularity_num_slow,
                              integrand_B63_singularity_den_slow,
                              integrand_B63_fast,integrand_B63_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B63 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B64
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B64_singularity_num_slow,
                              integrand_B64_singularity_den_slow,
                              integrand_B64_fast,integrand_B64_slow,
                              flag,rec_level,8,4,6,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B64 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B65
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B65_singularity_num_slow,
                              integrand_B65_singularity_den_slow,
                              integrand_B65_fast,integrand_B65_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B65 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B66
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B66_singularity_num_slow,
                              integrand_B66_singularity_den_slow,
                              integrand_B66_fast,integrand_B66_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B66 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B67
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B67_singularity_num_slow,
                              integrand_B67_singularity_den_slow,
                              integrand_B67_fast,integrand_B67_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B67 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 3x2 pieces (6 integrals). The last two,
// corresponding to [0.95,\pi] \times [-\delta,\delta] are split in a geometric
// way in the first coordinate, arithmetic in the second. The remaining four,
// corresponding to [\delta, 0.95] are split in an arithmetic way in both
// coordinates.

// This function computes the terms B_{71} to B_{76} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_7(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.65; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.65; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.65; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.95; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.65; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.95; right4[2] = 0;
  ivector left5(2); left5[1] = (interval)0.95; left5[2] = 0;
  ivector right5(2); right5[1] = p.Right; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.95; left6[2] = -p.Left;
  ivector right6(2); right6[1] = p.Right; right6[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
    
  // B71
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B71_singularity_num_slow,
                              integrand_B71_singularity_den_slow,
                              integrand_B71_fast,integrand_B71_slow,
                              flag,rec_level,6,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B71 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B72
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B72_singularity_num_slow,
                              integrand_B72_singularity_den_slow,
                              integrand_B72_fast,integrand_B72_slow,
                              flag,rec_level,4,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B72 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B73
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B73_singularity_num_slow,
                              integrand_B73_singularity_den_slow,
                              integrand_B73_fast,integrand_B73_slow,
                              flag,rec_level,8,4,6,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B73 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B74
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B74_singularity_num_slow,
                              integrand_B74_singularity_den_slow,
                              integrand_B74_fast,integrand_B74_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B74 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B75
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B75_singularity_num_slow,
                              integrand_B75_singularity_den_slow,
                              integrand_B75_fast,integrand_B75_slow,
                              flag,rec_level,5,4,4,4,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B75 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  // B76
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B76_singularity_num_slow,
                              integrand_B76_singularity_den_slow,
                              integrand_B76_fast,integrand_B76_slow,
                              flag,rec_level,5,8,4,8,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B76 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,0.65] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in an arithmetic way in both coordinates.

// This function computes part of the term B_{47} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_special_1(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.325; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.325; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.325; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.65; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.325; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.65; right4[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
    
  // B47 
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B47 part 1 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  return res2d_sing_second_var;
}



// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [0.65,0.95] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in an arithmetic way in both coordinates.

// This function computes part of the term B_{47} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_special_2(const ParameterSet& p, int& flag){
  ivector left5(2); left5[1] = (interval)0.65; left5[2] = 0;
  ivector right5(2); right5[1] = (interval)0.775; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.65; left6[2] = -p.Left;
  ivector right6(2); right6[1] = (interval)0.775; right6[2] = 0;
  ivector left7(2); left7[1] = (interval)0.775; left7[2] = 0;
  ivector right7(2); right7[1] = (interval)0.95; right7[2] = p.Left;
  ivector left8(2); left8[1] = (interval)0.775; left8[2] = -p.Left;
  ivector right8(2); right8[1] = (interval)0.95; right8[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
    
  // B47 
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left7,right7,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left8,right8,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B47 part 2 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  return res2d_sing_second_var;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [0.95,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in a geometric way in the first coordinate and in an arithmetic
// in the second.

// This function computes part of the term B_{47} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
interval compute_Lemma_222_singular_second_var_split_special_3(const ParameterSet& p, int& flag){
  ivector left9(2); left9[1] = (interval)0.95; left9[2] = 0;
  ivector right9(2); right9[1] = (interval)1.5; right9[2] = p.Left;
  ivector left10(2); left10[1] = (interval)0.95; left10[2] = -p.Left;
  ivector right10(2); right10[1] = (interval)1.5; right10[2] = 0;
  ivector left11(2); left11[1] = (interval)1.5; left11[2] = 0;
  ivector right11(2); right11[1] = p.Right; right11[2] = p.Left;
  ivector left12(2); left12[1] = (interval)1.5; left12[2] = -p.Left;
  ivector right12(2); right12[1] = p.Right; right12[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
    
  // B47 
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left9,right9,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left10,right10,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left11,right11,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left12,right12,integrate_singularity_y_axis_2D,
                              integrand_B47_singularity_num_slow,
                              integrand_B47_singularity_den_slow,
                              integrand_B47_fast,integrand_B47_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B47 part 3 " << curr_res << endl; res2d_sing_second_var += curr_res;
  
  return res2d_sing_second_var;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [\delta,0.65] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in an arithmetic way in both coordinates.

// This function computes part of the term B_{55} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
  interval compute_Lemma_222_singular_second_var_split_special_4(const ParameterSet& p, int& flag){
  ivector left1(2); left1[1] = p.Left; left1[2] = 0;
  ivector right1(2); right1[1] = (interval)0.325; right1[2] = p.Left;
  ivector left2(2); left2[1] = p.Left; left2[2] = -p.Left;
  ivector right2(2); right2[1] = (interval)0.325; right2[2] = 0;
  ivector left3(2); left3[1] = (interval)0.325; left3[2] = 0;
  ivector right3(2); right3[1] = (interval)0.65; right3[2] = p.Left;
  ivector left4(2); left4[1] = (interval)0.325; left4[2] = -p.Left;
  ivector right4(2); right4[1] = (interval)0.65; right4[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left1,right1,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left2,right2,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left3,right3,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++; 
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left4,right4,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B55 part 1 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [0.65,0.95] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in an arithmetic way in both coordinates.

// This function computes part of the term B_{55} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
  interval compute_Lemma_222_singular_second_var_split_special_5(const ParameterSet& p, int& flag){
  ivector left5(2); left5[1] = (interval)0.65; left5[2] = 0;
  ivector right5(2); right5[1] = (interval)0.775; right5[2] = p.Left;
  ivector left6(2); left6[1] = (interval)0.65; left6[2] = -p.Left;
  ivector right6(2); right6[1] = (interval)0.775; right6[2] = 0;
  ivector left7(2); left7[1] = (interval)0.775; left7[2] = 0;
  ivector right7(2); right7[1] = (interval)0.95; right7[2] = p.Left;
  ivector left8(2); left8[1] = (interval)0.775; left8[2] = -p.Left;
  ivector right8(2); right8[1] = (interval)0.95; right8[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D(p,left5,right5,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left6,right6,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left7,right7,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D(p,left8,right8,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B55 part 2 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}


// Computes \pa_t \pa_x z^{1}(0,0) for for A(\eps) = [1.08050,1.08055], 
// \eps = [0,1e-6] in the region [0.95,\pi] \times [-\delta,\delta]
// (singular second-coordinate). The integration is done by expanding in a Taylor series
// around (y,0) in the z-variable both numerator and denominator and cancelling 
// the resulting powers. If this fails, then we resort to Gauss-Legendre 
// of order 2.

// The domain of integration is split into 2x2 pieces (4 integrals), which are
// split in a geometric way in the first coordinate and in an arithmetic
// in the second.

// This function computes part of the term B_{55} (see paper).

// Input:
// * A ParameterSet p containing the necessary values of \eps, A(\eps) and the
// integration boundaries.

// Output:
// * An interval containing the calculated quantity.
// * An integer flag (passed by reference):
// - The last bit of flag is set to 1 if a division by zero is encountered.
// - The second to last bit of flag is set to 1 if we have reached the maximum depth in the recursion. 
// Otherwise flag equals zero.
  interval compute_Lemma_222_singular_second_var_split_special_6(const ParameterSet& p, int& flag){
  ivector left9(2); left9[1] = (interval)0.95; left9[2] = 0;
  ivector right9(2); right9[1] = (interval)1.5; right9[2] = p.Left;
  ivector left10(2); left10[1] = (interval)0.95; left10[2] = -p.Left;
  ivector right10(2); right10[1] = (interval)1.5; right10[2] = 0;
  ivector left11(2); left11[1] = (interval)1.5; left11[2] = 0;
  ivector right11(2); right11[1] = p.Right; right11[2] = p.Left;
  ivector left12(2); left12[1] = (interval)1.5; left12[2] = -p.Left;
  ivector right12(2); right12[1] = p.Right; right12[2] = 0;
  
  interval res2d_sing_second_var(0.0);
  interval curr_res(0.0);

  int rec_level = 0; rec_level++;
  
  // B55
  curr_res = (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left9,right9,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left10,right10,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left11,right11,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  curr_res+= (interval)2.0*(
             rec_integrate_singularity_2D_geometric(p,left12,right12,integrate_singularity_y_axis_2D,
                              integrand_B55_singularity_num_slow,
                              integrand_B55_singularity_den_slow,
                              integrand_B55_fast,integrand_B55_slow,
                              flag,rec_level,2,12,2,12,1));
  assert(rec_level == 0); if (flag & 1) return interval(0.0,0.0); rec_level++;
  cerr << "Z-axis singularity B55 part 3 " << curr_res << endl; res2d_sing_second_var += curr_res;
  return res2d_sing_second_var;
}

int main(int argc, char* argv[]){
  cout << SetPrecision(14,11);
  if (argc != 1) {
    cout << "Usage: " << endl;
    cout << argv[0] << endl;
    cout << "(and some patience...)" << endl;
    cout << "Results will be output to screen as soon as they are computed." << endl;
    exit(0);
  }
  
  ParameterSet PS_ini;
  // Neighbourhood of the singularity set to 2^(-6)
  "[0.03125,0.03125]" >> PS_ini.Left;
  
  // Boundary for the bounded part of the second interval
  PS_ini.Right = Pi_interval;
  PS_ini.abs_tol = PS_ini.rel_tol = 1e-4;
  
  "[1.08050,1.08055]" >> PS_ini.A;
  "[0.000000,0.000001]" >> PS_ini.epsilon;
  
  int flag = 0;

  #ifdef CASE_NONSINGULAR
    interval res = compute_Lemma_222_nonsingular(PS_ini, flag);
  #elif CASE_SINGULAR_CENTER
    interval res = compute_Lemma_222_singular_center(PS_ini, flag);
  #elif CASE_SINGULAR_FIRST
    interval res = compute_Lemma_222_singular_first_var(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_1
    interval res = compute_Lemma_222_singular_second_var_split_1(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_2
    interval res = compute_Lemma_222_singular_second_var_split_2(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_3
    interval res = compute_Lemma_222_singular_second_var_split_3(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_4
    interval res = compute_Lemma_222_singular_second_var_split_4(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_5
    interval res = compute_Lemma_222_singular_second_var_split_5(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_6
    interval res = compute_Lemma_222_singular_second_var_split_6(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_7
    interval res = compute_Lemma_222_singular_second_var_split_7(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_1
    interval res = compute_Lemma_222_singular_second_var_split_special_1(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_2
    interval res = compute_Lemma_222_singular_second_var_split_special_2(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_3
    interval res = compute_Lemma_222_singular_second_var_split_special_3(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_4
    interval res = compute_Lemma_222_singular_second_var_split_special_4(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_5
    interval res = compute_Lemma_222_singular_second_var_split_special_5(PS_ini,flag);
  #elif CASE_SINGULAR_SECOND_SPECIAL_6
    interval res = compute_Lemma_222_singular_second_var_split_special_6(PS_ini,flag);
  #else
    #error Wrong values of CASE, constant not set by preprocessor
  #endif 

  if ((flag&1)){ // critical
    cerr << "Critical case: division by zero encountered" << endl;
    return 0;
  }
  
  return 0;
}

